from pathlib import Path

import pytest

from i3configger import exc, partials

SCHEMES = Path(__file__).parents[1] / "examples" / "4-schemes" / "config.d"


def test_create():
    prts = partials.create(SCHEMES)
    assert len(prts) == 3


@pytest.mark.parametrize(
    "key, value, exp",
    (
        ("", "", exc.ConfigError),
        ("some-key", "", exc.ConfigError),
        ("", "", exc.ConfigError),
        (None, None, exc.ConfigError),
        ("non-existing-key", "some-value", exc.ConfigError),
        ("non-existing-key", "none-existing-value", exc.ConfigError),
        ("some-category", "none-existing-value", exc.ConfigError),
        ("some-key", "value1", True),
        ("some-key", "value2", True),
    ),
)
def test_select(key, value, exp):
    prts = partials.create(SCHEMES)
    selector = {key: value}
    if not isinstance(exp, bool):
        with pytest.raises(exp):
            partials.select(prts, selector)
    else:
        selected = partials.select(prts, selector)
        found = partials.find(prts, key, value)
        assert all(isinstance(p, partials.Partial) for p in selected)
        assert selected[0] == found
        assert selected[0].key == key
        assert selected[0].value == value



"""Inter Process Communication functionality."""
import logging
import subprocess
from pathlib import Path

log = logging.getLogger(__name__)


def configure(cnf=None, deactivate=False):
    if deactivate:
        I3.configure(None)
        I3bar.configure(None)
        Notify.configure(None)
        log.info("ipc deactivated")
    else:
        I3.configure(cnf.payload["main"]["i3_refresh_msg"])
        log.info(f"set i3 refresh method to {I3.refresh.__name__}")
        I3bar.configure(cnf.payload["main"]["status_command"])
        log.info(f"set i3bar refresh method to {I3bar.refresh.__name__}")
        Notify.configure(cnf.payload["main"]["notify"])
        log.info(f"set notify method to {Notify.send.__name__}")


def communicate(msg="new config active", refresh=False, urgency="low"):
    if refresh:
        I3.refresh()
        I3bar.refresh()
    Notify.send(msg, urgency=urgency)


class I3:
    @classmethod
    def configure(cls, which):
        cls.refresh = cls.METHOD_MAP.get(which, nop)

    @classmethod
    def reload_i3(cls):
        cls._send_i3_msg("reload")

    @classmethod
    def restart_i3(cls):
        subprocess.call(["i3-msg", "restart"])

    refresh = restart_i3
    METHOD_MAP = {"restart": restart_i3, "reload": reload_i3}

    @classmethod
    def _send_i3_msg(cls, msg):
        try:
            output = subprocess.check_output(["i3-msg", msg]).decode()
            if '"success":true' in output:
                return True
            return False
        except subprocess.CalledProcessError as e:
            if msg == "restart" and e.returncode == 1:
                log.debug("[IGNORE] exit 1 is ok for restart")
                return True

    @classmethod
    def get_config_error_report(cls, path):
        cmd = ["i3", "-C", "-c", str(path)]
        try:
            return subprocess.check_output(cmd).decode()
        except subprocess.CalledProcessError as e:
            return e.output.decode()
        except FileNotFoundError as e:
            assert Path(path).exists(), path
            assert "No such file or directory: 'i3'" in e.strerror
            log.warning("[IGNORE] crashed - no i3 -> assuming test system")
            return ""


class Notify:
    @classmethod
    def configure(cls, notify):
        cls.send = cls.notify_send if notify else nop

    @classmethod
    def notify_send(cls, msg, urgency="low"):
        """urgency levels: low, normal, critical"""
        subprocess.check_call(
            ["notify-send", "-a", "i3configger", "-t", "1", "-u", urgency, msg]
        )

    send = notify_send


class I3bar:
    command = "i3status"

    @classmethod
    def configure(cls, command):
        cls.command = command
        cls.refresh = cls.refresh if command else nop

    @classmethod
    def send_sigusr1(cls):
        try:
            subprocess.check_output(["killall", "-SIGUSR1", cls.command])
        except subprocess.CalledProcessError as e:
            log.debug("[IGNORE] failed status refresh: %s", e)

    refresh = send_sigusr1


def nop(*args, **kwargs):
    log.debug(f"just ignored everything to do with {args} and {kwargs}")



"""Functionality to create, find and select partial configurations."""
import logging
import pprint
import socket
from functools import total_ordering
from pathlib import Path
from typing import Union, List

from i3configger import base, exc

log = logging.getLogger(__name__)

SPECIAL_SELECTORS = {"hostname": socket.gethostname()}
EXCLUDE_MARKER = "."
"""config files starting with a dot are always excluded"""


@total_ordering
class Partial:
    def __init__(self, path: Path):
        self.path = path
        self.name = self.path.stem
        self.selectors = self.name.split(".")
        self.needsSelection = len(self.selectors) > 1
        self.key = self.selectors[0] if self.needsSelection else None
        self.value = self.selectors[1] if self.needsSelection else None
        self.lines = self.path.read_text().splitlines()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.path.name})"

    def __lt__(self, other):
        return self.name < other.name

    def get_pruned_content(self) -> str:
        """pruned content or '' if file only contains vars and comments"""
        lines = [l for l in self.lines if not l.strip().startswith(base.SET_MARK)]
        if not self.contain_something(lines):
            return ""
        while lines and not lines[0].strip():
            lines.pop(0)
        while lines and not lines[-1].strip():
            lines.pop()
        joinedLines = "\n".join(lines)
        return f"### {self.path.name} ###\n{joinedLines}\n\n"

    @staticmethod
    def contain_something(lines):
        for line in lines:
            line = line.strip()
            if not line:
                continue
            if line.startswith(base.COMMENT_MARK):
                continue
            if line.startswith(base.SET_MARK):
                continue
            return True

    @property
    def context(self):
        ctx = {}
        for line in [
            l.strip() for l in self.lines if l.strip().startswith(base.SET_MARK)
        ]:
            payload = line.split(maxsplit=1)[1]
            key, value = payload.split(maxsplit=1)
            ctx[key] = value
        return ctx


def find(
    prts: List[Partial], key: str, value: str = None
) -> Union[Partial, List[Partial]]:
    findings = []
    for prt in prts:
        if prt.key != key:
            continue
        if prt.value == value:
            return prt
        elif not value:
            findings.append(prt)
    return findings


def select(partials, selection, excludes=None) -> List[Partial]:
    def _select():
        selected.append(partial)
        if partial.needsSelection:
            del selection[partial.key]

    for key, value in SPECIAL_SELECTORS.items():
        if key not in selection:
            selection[key] = value
    selected: List[Partial] = []
    for partial in partials:
        if partial.needsSelection:
            if excludes and partial.key in excludes:
                log.debug(f"[IGNORE] {partial} (in {excludes})")
                continue
            if (
                selection
                and partial.key in selection
                and partial.value == selection.get(partial.key)
            ):
                _select()
        else:
            _select()
    log.debug(f"selected:\n{pprint.pformat(selected)}")
    if selection and not all(k in SPECIAL_SELECTORS for k in selection):
        raise exc.ConfigError(f"selection processed incompletely: {selection}")
    return selected


def create(partialsPath) -> List[Partial]:
    partialsPath = Path(partialsPath)
    assert partialsPath.is_dir(), partialsPath
    prts = []
    for path in partialsPath.glob(f"*{base.SUFFIX}"):
        if path.name.startswith(EXCLUDE_MARKER):
            log.info(f"excluding {path} because it starts with {EXCLUDE_MARKER}")
            continue
        prts.append(Partial(path))
    if not prts:
        raise exc.PartialsError(f"no '*{base.SUFFIX}' at {partialsPath}")
    return sorted(prts)



import json
from pathlib import Path

from i3configger import config
from i3configger.build import persist_results


def test_initialization(tmp_path, monkeypatch):
    """Given empty sources directory a new config is created from defaults"""
    monkeypatch.setattr(config, "get_i3wm_config_path", lambda: tmp_path)
    assert not (tmp_path / "config.d").exists()
    config.ensure_i3_configger_sanity()
    cnf = config.I3configgerConfig()
    assert cnf.configPath.exists()
    assert cnf.configPath.is_file()
    assert cnf.configPath.name == config.I3configgerConfig.CONFIG_NAME
    payload = json.loads(cnf.configPath.read_text())
    assert "main" in payload
    assert "bars" in payload
    assert "targets" in payload["bars"]
    assert "set" not in payload
    assert "select" not in payload
    assert "shadow" not in payload


def test_config_backup_is_not_overwritten(tmp_path):
    """Given an existing backup it is not overwritten by subsequent builds."""
    firstThing = "first thing"
    somePath = tmp_path / "some-path.txt"
    somePath.write_text(firstThing)
    persist_results({somePath: firstThing})
    assert (Path(str(somePath) + ".bak")).read_text() == firstThing
    somePath.write_text(firstThing)
    otherThing = "other thing"
    persist_results({somePath: otherThing})
    assert somePath.read_text() == otherThing
    assert (Path(str(somePath) + ".bak")).read_text() == firstThing



import logging
from functools import reduce
from string import Template
from typing import Iterable, Union

from i3configger import base, exc, partials

log = logging.getLogger(__name__)


def process(item: Iterable[Union[dict, partials.Partial]]) -> dict:
    ctx = merge_all(item)
    ctx = resolve_variables(ctx)
    ctx = remove_variable_markers(ctx)
    return ctx


def merge_all(item: Iterable[Union[dict, partials.Partial]]) -> dict:
    """Merge contexts of an arbitrary number of items."""
    dicts = []
    for elem in item:
        dicts.append(elem if isinstance(elem, dict) else elem.context)
    return reduce(merge, dicts)


def merge(dst: dict, src: dict) -> dict:
    """merge source into destination mapping (overwrite existing keys)."""
    for key in src:
        if key in dst:
            if isinstance(dst[key], dict) and isinstance(src[key], dict):
                merge(dst[key], src[key])
            else:
                dst[key] = src[key]
        else:
            dst[key] = src[key]
    return dst


def prune(dst: dict, src: dict) -> dict:
    for key, value in src.items():
        if isinstance(value, dict):
            prune(dst.get(key, {}), value)
        elif key in dst:
            del dst[key]
    return dst


def resolve_variables(ctx: dict) -> dict:
    """If variables are set by a variable, replace them by their value."""
    while any(v.startswith(base.VAR_MARK) for v in ctx.values()):
        for key, value in ctx.items():
            if value.startswith(base.VAR_MARK):
                ctx[key] = _resolve_variable(value, ctx, path=key)
    return ctx


def _resolve_variable(key, ctx, path):
    path += "->" + key
    resolved = ctx.get(key)
    if not resolved:
        raise exc.ContextError(f"not resolvable: {path}")
    if resolved.startswith(base.VAR_MARK):
        return _resolve_variable(resolved, ctx, path)
    return resolved


def remove_variable_markers(ctx: dict) -> dict:
    cleaned = {}
    lvm = len(base.VAR_MARK)
    for key, value in ctx.items():
        key = key[lvm:] if key.startswith(base.VAR_MARK) else key
        cleaned[key] = value
    return cleaned


def substitute(content: str, ctx: dict):
    """Substitute all variables with their values.

    Works out of the box, because '$' is the standard substitution
    marker for string.Template

    As there also might be other occurrences of "$" (e.g. in regexes)
    `safe_substitute()` is used,
    """
    return Template(content).safe_substitute(ctx)



"""Custom errors."""


class I3configgerException(Exception):
    """Main exception with log style string formatting enhancement"""


class BuildError(I3configgerException):
    pass


class ConfigError(I3configgerException):
    pass


class MessageError(I3configgerException):
    pass


class PartialsError(I3configgerException):
    pass


class ParseError(I3configgerException):
    pass


class ContextError(I3configgerException):
    pass


class UserError(I3configgerException):
    """User did something wrong and I can't go on."""



"""Functionality to watch the configuration dir to trigger rebuilds on changes."""
import time

import logging
import os
import sys
from pathlib import Path

import daemon  # type: ignore
import psutil  # type: ignore

from i3configger import base, build, exc, ipc, config
from i3configger.inotify_simple import INotify, flags

log = logging.getLogger(__name__)


_MASK = (
    flags.CREATE
    | flags.ATTRIB
    | flags.DELETE
    | flags.MODIFY
    | flags.CLOSE_WRITE
    | flags.MOVED_FROM
    | flags.MOVED_TO
    | flags.DELETE_SELF
    | flags.MOVE_SELF
)
"""Tell inotify to trigger on changes"""


def watch_guarded():
    while True:
        try:
            watch_unguarded()
        except exc.I3configgerException as e:
            ipc.communicate("WARNING", urgency="normal")
            log.warning(str(e))
        except Exception as e:
            ipc.communicate("ERROR", urgency="critical")
            log.error(str(e))


def watch_unguarded():
    cnf = config.I3configgerConfig(load=False)
    watcher = INotify()
    watcher.add_watch(str(cnf.partialsPath).encode(), mask=_MASK)
    log.debug(f"start watching {cnf.partialsPath}")
    while True:
        events = watcher.read(read_delay=50)
        log.debug(f"events: {[f'{e[3]}:m={e[1]}' for e in events]}")
        if not events:
            continue
        paths = [cnf.partialsPath / e[-1] for e in events]
        if any(p.suffix in [base.SUFFIX, ".json"] for p in paths):
            build.build_all()
            ipc.communicate(refresh=True)


def get_i3configger_process():
    """should always be max one, but you never know ..."""
    all_ = [p for p in psutil.process_iter() if p.name() == "i3configger"]
    others = [p for p in all_ if p.pid != os.getpid()]
    if len(others) == 1:
        return others[0]
    elif len(others) > 1:
        raise exc.I3configgerException(
            f"More than one i3configger running: {others}"
            f"If this happens again, please file an issue "
            f"and tell me how you did it."
        )


def daemonized(verbosity, logPath):
    process = get_i3configger_process()
    if process:
        raise exc.UserError(f"i3configger already running ({process})")
    context = daemon.DaemonContext(working_directory=Path(__file__).parent)
    if verbosity > 2:
        # spew output to terminal from where daemon was started
        context.stdout = sys.stdout
        context.stderr = sys.stderr
    with context:
        base.configure_logging(verbosity, logPath, isDaemon=True)
        watch_guarded()


def exorcise():
    process = get_i3configger_process()
    if not process:
        raise exc.UserError("no daemon running - nothing to kill")
    process.kill()
    # sometimes the process needs a bit longer to die
    for _ in range(20):
        if not process.is_running():
            break
        time.sleep(0.1)
    else:
        raise exc.I3configgerException(f"process {process} does not want to die")
    log.info(f"killed {process}")



import pytest as pytest

from i3configger import context, exc

verySimple = {"$k1": "v1", "$k2": "v2"}
simple = {"$k1": "v1", "$k2": "v1"}
undef = {"$k1": "v1", "$k2": "$undef"}
oneIndirection = {"$k1": "$k2", "$k2": "v1"}
twoIndirections = {"$k1": "$k2", "$k2": "$k3", "$k3": "v1"}
unresolvedIndirection = {"$k1": "$k2", "$k2": "$k3", "$k3": "$k4"}


@pytest.mark.parametrize(
    "ctx, exp",
    (
        (verySimple, verySimple),
        (simple, {"$k1": "v1", "$k2": "v1"}),
        (undef, exc.ContextError),
        (oneIndirection, {"$k1": "v1", "$k2": "v1"}),
        (twoIndirections, {"$k1": "v1", "$k2": "v1", "$k3": "v1"}),
        (unresolvedIndirection, exc.ContextError),
    ),
)
def test_context(ctx, exp):
    if isinstance(exp, dict):
        ctx = context.resolve_variables(ctx)
        assert ctx == exp
    else:
        with pytest.raises(exc.ContextError):
            context.resolve_variables(ctx)



"""i3configger configuration functionality."""
import copy
import json
import logging
import os
import pprint
import shutil
from pathlib import Path

from i3configger import context, exc

log = logging.getLogger(__name__)
cliMainOverrideMap: dict = {}
"""holds parsed args if started from cli and was initialized"""

DEFAULTS = {
    "main": {
        "target": "../config",
        "i3_refresh_msg": "reload",
        "status_command": "i3status",
        "log": None,
        "notify": False,
    },
    "bars": {
        "defaults": {"template": "tpl", "target": "..", "select": "default"},
        "targets": {},
    },
}
CONFIG_CANDIDATES = [
    Path("~/.i3").expanduser(),
    Path(os.getenv("XDG_CONFIG_HOME", "~/.config/")).expanduser() / "i3",
]


class I3configgerConfig:
    PARTIALS_NAME = "config.d"
    CONFIG_NAME = "i3configger.json"
    MESSAGES_NAME = ".messages.json"

    def __init__(self, load=True):
        i3configBasePath = get_i3wm_config_path()
        self.partialsPath = i3configBasePath / self.PARTIALS_NAME
        self.configPath = self.partialsPath / self.CONFIG_NAME
        self.messagesPath = self.partialsPath / self.MESSAGES_NAME
        if load:
            self.load()

    def __str__(self):
        return f"{self.__class__.__name__}:\n{pprint.pformat(vars(self))}"

    def load(self):
        """Layered overrides `DEFAULTS` -> `i3configger.json`-> `args`"""
        cnfFromDefaults = copy.deepcopy(DEFAULTS)
        cnfFromFile = fetch(self.configPath)
        self.payload = context.merge(cnfFromDefaults, cnfFromFile)
        mainOverrides = dict(main=cliMainOverrideMap)
        self.payload = context.merge(self.payload, mainOverrides)
        targetPath = Path(self.payload["main"]["target"]).expanduser()
        if targetPath.is_absolute():
            self.targetPath = targetPath.resolve()
        else:
            self.targetPath = (self.partialsPath / targetPath).resolve()

    def get_bar_targets(self):
        """Create a resolved copy of the bar settings."""
        barTargets = {}
        barSettings = self.payload.get("bars", {})
        if not barSettings:
            return barTargets
        defaults = barSettings.get("defaults", {})
        for name, bar in barSettings["targets"].items():
            newBar = dict(bar)
            barTargets[name] = newBar
            for defaultKey, defaultValue in defaults.items():
                if defaultKey not in newBar:
                    newBar[defaultKey] = defaultValue
            container = Path(newBar["target"])
            if not container.is_absolute():
                container = (self.partialsPath / container).resolve()
            newBar["target"] = str(container)
        return barTargets


def fetch(path: Path) -> dict:
    if not path.exists():
        raise exc.ConfigError(f"config not found at {path}")
    payload = json.loads(path.read_text())
    log.debug(f"{path}:\n{pprint.pformat(payload)}")
    return payload


def freeze(path, obj):
    path.write_text(json.dumps(obj, sort_keys=True, indent=2))
    log.debug(f"froze {pprint.pformat(obj)} to {path}")


def ensure_i3_configger_sanity():
    i3wmConfigPath = get_i3wm_config_path()
    partialsPath = i3wmConfigPath / I3configgerConfig.PARTIALS_NAME
    if not partialsPath.exists():
        log.info(f"create new config folder at {partialsPath}")
        partialsPath.mkdir()
        for candidate in [i3wmConfigPath / "config", Path("etc/i3/config")]:
            if candidate.exists():
                log.info(f"populate config with {candidate}")
                shutil.copy2(candidate, partialsPath / "config.conf")
    configPath = partialsPath / I3configgerConfig.CONFIG_NAME
    if not configPath.exists():
        log.info(f"create default configuration at {configPath}")
        freeze(configPath, DEFAULTS)


def get_i3wm_config_path():
    """Use same search order like i3 (no system stuff though).

    see: https://github.com/i3/i3/blob/4.13/libi3/get_config_path.c#L31
    """
    for candidate in CONFIG_CANDIDATES:
        if candidate.exists():
            return candidate
    raise exc.ConfigError(
        f"can't find i3 config at the standard locations: {CONFIG_CANDIDATES}"
    )



"""Functionality implementing the messaging mechanism."""
import logging
from pathlib import Path

from i3configger import base, config, context, exc, partials

log = logging.getLogger(__name__)


class CMD:
    MERGE = "merge"
    PRUNE = "prune"
    SELECT = "select"
    SELECT_NEXT = "select-next"
    SELECT_PREVIOUS = "select-previous"
    SET = "set"
    SHADOW = "shadow"

    @classmethod
    def get_all_commands(cls):
        return [v for k, v in cls.__dict__.items() if k[0].isupper() and k[0] != "_"]


def save(message):
    cnf = config.I3configgerConfig(load=False)
    prts = partials.create(cnf.partialsPath)
    messenger = Messenger(cnf.messagesPath, prts, message)
    messenger.digest_message()
    config.freeze(cnf.messagesPath, messenger.payload)


class Messenger:
    def __init__(self, messagesPath, prts, message=None):
        self.messagesPath = messagesPath
        self.prts = prts
        self.message = message
        if message:
            if len(message) < 2:
                raise exc.UserError(f"message needs at least key and value ({message})")
            self.command, self.key, *rest = message
            self.value = rest[0] if rest else ""
            if self.command != CMD.SHADOW and ":" in self.key:
                raise exc.UserError(f"nesting of keys only sensible with {CMD.SHADOW}")
        self.payload = self.fetch_messages()
        log.debug(f"send message '{message}' to {messagesPath}")

    def digest_message(self):
        try:
            self.COMMAND_METHOD_MAP[self.command](self)
        except KeyError:
            raise exc.UserError(
                f"Unknown command: {self.command}. "
                f"Use one of {', '.join(CMD.get_all_commands())}"
            )

    def _process_merge(self):
        self._transform(context.merge)

    def _process_prune(self):
        self._transform(context.prune)

    def _transform(self, func):
        path = Path(self.key).expanduser()
        if not path.is_absolute():
            path = self.messagesPath.parent / path
        self.payload = func(self.payload, config.fetch(path))

    def _process_set(self):
        if self.value.lower() == base.DEL:
            del self.payload[CMD.SET][base.VAR_MARK + self.key]
        else:
            self.payload[CMD.SET][base.VAR_MARK + self.key] = self.value

    def _process_select(self):
        candidates = partials.find(self.prts, self.key)
        if not candidates:
            raise exc.MessageError(f"No candidates for {self.message} in {self.prts}")
        candidate = partials.find(self.prts, self.key, self.value)
        if not candidate:
            raise exc.MessageError(f"No candidates for {self.message} in {candidates}")
        if self.value and self.value.lower() == base.DEL:
            del self.payload[CMD.SELECT][self.key]
        else:
            self.payload[CMD.SELECT][self.key] = candidate.value

    def _process_shadow(self):
        """Shadow arbitrary settings made in i3configger.json.

        key:deeper:deepest[...] -> [key][deeper][deepest][...]
        """
        parts = self.key.split(":")
        current = self.payload[CMD.SHADOW]
        while True:
            part = parts.pop(0)
            if parts:
                current[part] = {}
                current = current[part]
            else:
                if self.value is not None and self.value.lower() == base.DEL:
                    del current[part]
                else:
                    current[part] = self.value
                break

    def _process_select_shift(self):
        candidates = partials.find(self.prts, self.key)
        if not candidates:
            raise exc.MessageError(f"No candidates for {self.message} in {self.prts}")
        if self.command == CMD.SELECT_PREVIOUS:
            candidates = list(reversed(candidates))
        current = self.payload["select"].get(self.key) or candidates[0].key
        for idx, candidate in enumerate(candidates):
            if candidate.value == current:
                try:
                    new = candidates[idx + 1]
                except IndexError:
                    new = candidates[0]
                break
        else:
            new = candidates[0]
        log.info(f"select {self.key}.{new}")
        self.payload[CMD.SELECT][self.key] = new.value

    def fetch_messages(self):
        if self.messagesPath.exists():
            messages = config.fetch(self.messagesPath)
        else:
            messages = {}
        self.ensure_message_keys(messages, self.prts)
        return messages

    def ensure_message_keys(self, state, prts):
        if CMD.SELECT not in state:
            initialSelects = {}
            for prt in prts:
                if not prt.needsSelection:
                    continue
                if prt.key not in initialSelects and prt.key != base.I3BAR:
                    initialSelects[prt.key] = prt.value
            state[CMD.SELECT] = initialSelects
        if CMD.SET not in state:
            state[CMD.SET] = {}
        if CMD.SHADOW not in state:
            state[CMD.SHADOW] = {}

    COMMAND_METHOD_MAP = {
        CMD.MERGE: _process_merge,
        CMD.PRUNE: _process_prune,
        CMD.SELECT: _process_select,
        CMD.SELECT_NEXT: _process_select_shift,
        CMD.SELECT_PREVIOUS: _process_select_shift,
        CMD.SET: _process_set,
        CMD.SHADOW: _process_shadow,
    }



from pathlib import Path

import pytest

from i3configger import build, config, ipc

HERE = Path(__file__).parent
EXAMPLES = HERE.parent / "examples"
REFERENCE = HERE / "examples"
FAKE_HEADER = """\
###############################################################################
# Built by i3configger /from/some/directory/who/cares  (some time after 1972) #
###############################################################################
"""
TEST_FOLDER_NAMES = sorted(
    list(
        [
            d.name
            for d in EXAMPLES.iterdir()
            if d.is_dir() and not str(d.name).startswith("_")
        ]
    )
)


@pytest.mark.parametrize("container", TEST_FOLDER_NAMES)
def test_build(container, monkeypatch):
    ipc.configure(deactivate=True)
    monkeypatch.setattr(config, "get_i3wm_config_path", lambda: EXAMPLES / container)
    monkeypatch.setattr(build, "make_header", lambda _: FAKE_HEADER)
    monkeypatch.setattr(build, "check_config", lambda _: True)
    config.ensure_i3_configger_sanity()
    cnf = config.I3configgerConfig()
    assert cnf.configPath.exists() and cnf.configPath.is_file()
    build.build_all()
    buildPath = cnf.configPath.parents[1]
    referencePath = REFERENCE / container
    names = [p.name for p in referencePath.iterdir()]
    assert names
    for name in names:
        resultFilePath = buildPath / name
        referenceFilePath = referencePath / name
        assert resultFilePath != referenceFilePath
        result = resultFilePath.read_text()
        reference = referenceFilePath.read_text()
        assert result == reference



import shutil
import subprocess
from contextlib import suppress

import pytest

from i3configger import exc, watch, config


class Runner:
    COMMAND_NAME = "i3configger"

    def __init__(self, cwd):
        self.cwd = cwd

    def __call__(self, args=None, otherCwd=None):
        cmd = [self.COMMAND_NAME]
        if args:
            cmd += args
        cp = subprocess.run(
            cmd,
            timeout=1,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=otherCwd or self.cwd,
        )
        return cp.returncode, cp.stdout.decode(), cp.stderr.decode()


@pytest.fixture(scope="session", name="configPath", autouse=True)
def ensure_config_path_exists():
    try:
        yield config.get_i3wm_config_path()
    except exc.ConfigError:
        path = config.CONFIG_CANDIDATES[0]
        path.mkdir()
        yield path
        shutil.rmtree(path, ignore_errors=True)


@pytest.fixture(name="runner")
def create_i3configger_runner(monkeypatch, tmp_path) -> Runner:
    with suppress(exc.UserError):
        watch.exorcise()
    assert not watch.get_i3configger_process()
    monkeypatch.chdir(tmp_path)
    return Runner(tmp_path)


def test_help(runner):
    ret, out, err = runner(["--help"])
    assert ret == 0
    assert not err
    assert "usage:" in out


def test_daemon(runner):
    ret, out, err = runner(["--daemon"])
    assert ret == 0
    assert not err
    assert not out
    assert watch.get_i3configger_process()
    watch.exorcise()
    assert not watch.get_i3configger_process()


def test_kill_with_running_i3configger_works(runner):
    runner(["--daemon"])
    assert watch.get_i3configger_process()
    ret, out, err = runner(["--kill"])
    assert ret == 0
    assert not err
    assert not out
    assert not watch.get_i3configger_process()


def test_kill_without_running_i3configger_gives_good_error(runner):
    ret, out, err = runner(["--kill"])
    assert ret == 1
    assert "no daemon running - nothing to kill" in err
    assert not out
    assert not watch.get_i3configger_process()



from pathlib import Path

from setuptools import find_packages, setup


def make_long_description():
    here = Path(__file__).parent
    readme = (here / "README.md").read_text()
    changelog = (here / "CHANGELOG.md").read_text()
    return f"{readme}\n\n{changelog}"


kwargs = dict(
    name="i3configger",
    author="Oliver Bestwalter",
    url="https://github.com/obestwalter/i3configger",
    description="i3 config manipulation tool",
    long_description=make_long_description(),
    long_description_content_type="text/markdown",
    use_scm_version=True,
    python_requires=">=3.6",
    setup_requires=["setuptools_scm"],
    entry_points={"console_scripts": ["i3configger = i3configger.cli:main"]},
    install_requires=["psutil", "python-daemon"],
    extras_require={
        "lint": ["pre-commit"],
        "test": ["pytest"],
        "docs": ["mkdocs", "mkdocs-material"],
        "release": ["plumbum", "twine", "readme_renderer[md]"],
    },
    packages=find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX :: Linux",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Topic :: Utilities",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)


if __name__ == "__main__":
    setup(**kwargs)



"""WARNING Just an experiment - please ignore this."""
from i3configger import config

BINDCODE = "bindcode"
BINDSYM = "bindsym"


class Bindings:
    """
    bindsym | bindcode
    [--release] [<Group>+][<Modifiers>+]<keysym> command

    [--release] [--border] [--whole-window] [<Modifiers>+]button<n> command
    """

    def __init__(self, content):
        self.content = content

    def get_all_bindings(self):
        lines = [l.strip() for l in self.content.splitlines()]
        lines = [l for l in lines if any(m in l for m in [BINDCODE, BINDSYM])]
        lines = [l for l in lines if not l.startswith(base.COMMENT_MARK)]
        return sorted(set(lines))

    def translate_bindings(self):
        """translate bindcode to bindsym assignments

        this need to be done the moment the information is asked because it
        depends on the currently active layout.
        """
        raise NotImplementedError()

    def write_bindings_info(self):
        """Write info in some format that can be nicely displayed"""
        raise NotImplementedError()


if __name__ == "__main__":
    # TODO use partials and account for modes
    # a naming convention would make this quite easy
    # mode-<modename>.conf -> bindings active in <modename>
    from i3configger import base

    p = config.I3configgerConfig().targetPath
    b = Bindings(p.read_text())
    print("\n".join(b.get_all_bindings()))



"""Vendored version of inotify_simple as suggested by author
https://github.com/chrisjbillington/inotify_simple

Copyright (c) 2016, Chris Billington
Copyright (c) 2018, Oliver Bestwalter (simplifications/Py3 only)

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided wi6h the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import enum
import collections
import struct
import select
import time
import ctypes
from os import strerror, fsencode, read, close, fsdecode
from errno import EINTR
from termios import FIONREAD
from fcntl import ioctl


__all__ = ["flags", "masks", "parse_events", "INotify", "Event"]
_libc = ctypes.cdll.LoadLibrary("libc.so.6")
_libc.__errno_location.restype = ctypes.POINTER(ctypes.c_int)

Event = collections.namedtuple("Event", ["wd", "mask", "cookie", "name"])
"""A ``namedtuple`` (wd, mask, cookie, name) for an inotify event.

``namedtuple`` objects are very lightweight to instantiate and access,
whilst being human readable when printed, which is useful for debugging and logging.
For best performance, note that element access by index is about four times faster than by name.
"""
_EVENT_STRUCT_FORMAT = "iIII"
_EVENT_STRUCT_SIZE = struct.calcsize(_EVENT_STRUCT_FORMAT)


def _libc_call(func, *args):
    """Wrapper which raises errors and retries on EINTR."""
    while True:
        rc = func(*args)
        if rc == -1:
            errno = _libc.__errno_location().contents.value
            if errno == EINTR:
                # retry
                continue
            else:
                raise OSError(errno, strerror(errno))
        return rc


class INotify(object):
    def __init__(self):
        """Object wrapper around ``inotify_init()`` which stores the inotify file
        descriptor. Raises an OSError on failure. :func:`~inotify_simple.INotify.close`
        should be called when no longer needed. Can be used as a context manager
        to ensure it is closed."""
        #: The inotify file descriptor returned by ``inotify_init()``. You are
        #: free to use it directly with ``os.read`` if you'd prefer not to call
        #: :func:`~inotify_simple.INotify.read` for some reason.
        self.fd = _libc_call(_libc.inotify_init)
        self._poller = select.poll()
        self._poller.register(self.fd)

    def add_watch(self, path, mask):
        """Wrapper around ``inotify_add_watch()``. Returns the watch
        descriptor or raises an OSError on failure.

        Args:
            path (py3 str or bytes, py2 unicode or str): The path to watch.
                If ``str`` in python3 or ``unicode`` in python2, will be encoded with
                the filesystem encoding before being passed to
                ``inotify_add_watch()``. Note that ``pathlib.Path`` objects are
                sufficiently string-like to be passed to this method as-is.

            mask (int): The mask of events to watch for. Can be constructed by
                bitwise-ORing :class:`~inotify_simple.flags` together.

        Returns:
            int: watch descriptor"""
        if not isinstance(path, bytes):
            path = fsencode(path)
        return _libc_call(_libc.inotify_add_watch, self.fd, path, mask)

    def rm_watch(self, wd):
        """Wrapper around ``inotify_rm_watch()``. Raises OSError on failure.

        Args:
            wd (int): The watch descriptor to remove"""
        _libc_call(_libc.inotify_rm_watch, self.fd, wd)

    def read(self, timeout=None, read_delay=None):
        """Read the inotify file descriptor and return the resulting list of
        :attr:`~inotify_simple.Event` namedtuples (wd, mask, cookie, name).

        Args:
            timeout (int): The time in milliseconds to wait for events if
                there are none. If `negative or `None``, block until there are
                events.

            read_delay (int): The time in milliseconds to wait after the first
                event arrives before reading the buffer. This allows further
                events to accumulate before reading, which allows the kernel
                to consolidate like events and can enhance performance when
                there are many similar events.

        Returns:
            list: list of :attr:`~inotify_simple.Event` namedtuples"""
        # Wait for the first event:
        pending = self._poller.poll(timeout)
        if not pending:
            # Timed out, no events
            return []
        if read_delay is not None:
            # Wait for more events to accumulate:
            time.sleep(read_delay / 1000.0)
        # How much data is available to read?
        bytes_avail = ctypes.c_int()
        ioctl(self.fd, FIONREAD, bytes_avail)
        buffer_size = bytes_avail.value
        # Read and parse it:
        data = read(self.fd, buffer_size)
        events = parse_events(data)
        return events

    def close(self):
        """Close the inotify file descriptor"""
        close(self.fd)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


def parse_events(data):
    """Parse data read from an inotify file descriptor into list of
    :attr:`~inotify_simple.Event` namedtuples (wd, mask, cookie, name). This
    function can be used if you have decided to call ``os.read()`` on the
    inotify file descriptor yourself, instead of calling
    :func:`~inotify_simple.INotify.read`.

    Args:
        data (bytes): A byte string as read from an inotify file descriptor
    Returns:
        list: list of :attr:`~inotify_simple.Event` namedtuples"""
    events = []
    offset = 0
    buffer_size = len(data)
    while offset < buffer_size:
        wd, mask, cookie, namesize = struct.unpack_from(
            _EVENT_STRUCT_FORMAT, data, offset
        )
        offset += _EVENT_STRUCT_SIZE
        name = fsdecode(
            ctypes.c_buffer(data[offset : offset + namesize], namesize).value
        )
        offset += namesize
        events.append(Event(wd, mask, cookie, name))
    return events


class flags(enum.IntEnum):
    """Inotify flags as defined in ``inotify.h`` but with ``IN_`` prefix
    omitted. Includes a convenience method for extracting flags from a mask.
    """

    ACCESS = 0x00000001  #: File was accessed
    MODIFY = 0x00000002  #: File was modified
    ATTRIB = 0x00000004  #: Metadata changed
    CLOSE_WRITE = 0x00000008  #: Writable file was closed
    CLOSE_NOWRITE = 0x00000010  #: Unwritable file closed
    OPEN = 0x00000020  #: File was opened
    MOVED_FROM = 0x00000040  #: File was moved from X
    MOVED_TO = 0x00000080  #: File was moved to Y
    CREATE = 0x00000100  #: Subfile was created
    DELETE = 0x00000200  #: Subfile was deleted
    DELETE_SELF = 0x00000400  #: Self was deleted
    MOVE_SELF = 0x00000800  #: Self was moved

    UNMOUNT = 0x00002000  #: Backing fs was unmounted
    Q_OVERFLOW = 0x00004000  #: Event queue overflowed
    IGNORED = 0x00008000  #: File was ignored

    ONLYDIR = 0x01000000  #: only watch the path if it is a directory
    DONT_FOLLOW = 0x02000000  #: don't follow a sym link
    EXCL_UNLINK = 0x04000000  #: exclude events on unlinked objects
    MASK_ADD = 0x20000000  #: add to the mask of an already existing watch
    ISDIR = 0x40000000  #: event occurred against dir
    ONESHOT = 0x80000000  #: only send event once

    @classmethod
    def from_mask(cls, mask):
        """Convenience method. Return a list of every flag in a mask."""
        return [flag for flag in cls.__members__.values() if flag & mask]


class masks(enum.IntEnum):
    """Convenience masks as defined in ``inotify.h`` but with ``IN_`` prefix
    omitted."""

    #: helper event mask equal to ``flags.CLOSE_WRITE | flags.CLOSE_NOWRITE``
    CLOSE = flags.CLOSE_WRITE | flags.CLOSE_NOWRITE
    #: helper event mask equal to ``flags.MOVED_FROM | flags.MOVED_TO``
    MOVE = flags.MOVED_FROM | flags.MOVED_TO

    #: bitwise-OR of all the events that can be passed to
    #: :func:`~inotify_simple.INotify.add_watch`
    ALL_EVENTS = (
        flags.ACCESS
        | flags.MODIFY
        | flags.ATTRIB
        | flags.CLOSE_WRITE
        | flags.CLOSE_NOWRITE
        | flags.OPEN
        | flags.MOVED_FROM
        | flags.MOVED_TO
        | flags.DELETE
        | flags.CREATE
        | flags.DELETE_SELF
        | flags.MOVE_SELF
    )



"""Command Line Interface and main entry point."""
import argparse
import logging
import sys

from i3configger import base, build, config, exc, ipc, message, watch

log = logging.getLogger(__name__)


def main():
    """Wrap main to show own exceptions wo traceback in normal use."""
    args = process_command_line()
    try:
        _main(args)
    except exc.I3configgerException as e:
        if args.v > 2:
            raise
        sys.exit(f"[FATAL] {e}")


def _main(args):
    config.ensure_i3_configger_sanity()
    cnf = config.I3configgerConfig()
    ipc.configure(cnf)
    base.configure_logging(verbosity=args.v, logPath=cnf.payload["main"]["log"])
    if args.version:
        print(f"i3configger {base.get_version()}")
        return 0

    if args.kill:
        watch.exorcise()
        return 0

    if args.message:
        message.save(args.message)
    if watch.get_i3configger_process():
        if not args.message:
            raise exc.UserError("already running - did you mean to send a message?")
        log.info(f"{args.message} is saved - now let the running process do the work")
        return 0

    if args.daemon:
        watch.daemonized(args.v, args.log)
    elif args.watch:
        try:
            watch.watch_guarded()
        except KeyboardInterrupt:
            sys.exit("interrupted by user")
    else:
        build.build_all()
        ipc.communicate(refresh=True)


def process_command_line():
    parser = argparse.ArgumentParser(
        "i3configger", formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    args = _parse_args(parser)
    config.cliMainOverrideMap = vars(args)
    if args.message and any([args.daemon, args.kill, args.watch]):
        parser.error(
            "message and daemon/watch can't be used together. "
            "Start the watcher process first and then you can send "
            "messages in following calls."
        )
    return args


def _parse_args(p):
    """Command line arguments - all optional with [reasonable] defaults"""
    p.add_argument(
        "--version", action="store_true", help="show version information and exit"
    )
    p.add_argument("-v", action="count", help="raise verbosity", default=0)
    g = p.add_mutually_exclusive_group()
    g.add_argument(
        "--watch",
        action="store_true",
        help="watch and build in foreground",
        default=False,
    )
    g.add_argument(
        "--daemon", action="store_true", help="watch and build as daemon", default=False
    )
    g.add_argument(
        "--kill", action="store_true", default=False, help="exorcise daemon if running"
    )
    p.add_argument(
        "--i3-refresh-msg",
        action="store",
        default="reload",
        choices=["restart", "reload", "nop"],
        help="i3-msg to send after build",
    )
    p.add_argument(
        "--notify",
        action="store_true",
        default=False,
        help="show build notification via notify-send",
    )
    p.add_argument(
        "--log",
        action="store",
        default=None,
        help="i3configgerPath to where log should be stored",
    )
    p.add_argument("message", help="message to send to i3configger", nargs="*")
    return p.parse_args()



"""Basic names and functionality."""
import logging
import os
import sys
import tempfile
from pathlib import Path

from i3configger import exc

log = logging.getLogger(__name__)
DEBUG = os.getenv("DEBUG", False)
COMMENT_MARK = "#"
VAR_MARK = "$"
SET_MARK = "set"
SUFFIX = ".conf"
I3BAR = "i3bar"
"""reserved key for status bar template files"""
DEL = "del"
"""signal to delete a key in shadow or set"""


def configure_logging(verbosity: int, logPath, isDaemon=False):
    rootLogger = logging.getLogger()
    verbosity = 3 if DEBUG else verbosity
    level = {0: "ERROR", 1: "WARNING", 2: "INFO"}.get(verbosity, "DEBUG")
    if logPath:
        logPath = Path(logPath).expanduser()
    else:
        name = "i3configger-daemon.log" if isDaemon else "i3configger.log"
        logPath = Path(tempfile.gettempdir()) / name
        if verbosity > 1:
            print(f"logging to {logPath} with level {level}", file=sys.stderr)
    fmt = "%(asctime)s %(name)s:%(funcName)s:%(lineno)s %(levelname)s: %(message)s"
    if not rootLogger.handlers:
        logging.basicConfig(format=fmt, level=level)
    fileHandler = logging.FileHandler(logPath)
    fileHandler.setFormatter(logging.Formatter(fmt))
    fileHandler.setLevel(level)
    rootLogger.addHandler(fileHandler)


def get_version():
    """hide behind a wrapped function (slow and not a catastrophe if fails)"""
    try:
        from pkg_resources import get_distribution

        return get_distribution("i3configger").version
    except Exception:
        log.exception("fetching version failed")
        return "unknown"


def i3configger_excepthook(type_, value, traceback):
    """Make all exceptions look like a friendly error message :)"""
    if DEBUG or not isinstance(value, exc.I3configgerException):
        _REAL_EXCEPTHOOK(type_, value, traceback)
    else:
        sys.exit(f"[FATAL] {type(value).__name__}: {value}")


_REAL_EXCEPTHOOK = sys.excepthook
sys.excepthook = i3configger_excepthook



import sys

import logging
import shutil
from pathlib import Path
from plumbum import local, ProcessExecutionError
from twine.commands.check import check as twine_check
from twine.commands.upload import upload as twine_upload
from twine.settings import Settings as twine_Settings

log = logging.getLogger(__name__)
PROJECT_ROOT_PATH = Path(__file__).parent
DIST_PATH = PROJECT_ROOT_PATH / "dist"
git = local["git"]


def main():
    logging.basicConfig(level=logging.DEBUG)
    if repo_is_dirty():
        raise EnvironmentError("repo is dirty!")
    if len(sys.argv) < 2:
        raise ValueError("need a version!")
    version = sys.argv[1]
    dryRun = len(sys.argv) > 2
    with local.cwd(PROJECT_ROOT_PATH):
        release(version, dryRun)


def release(version, dryRun):
    tidy_up()
    tag_repo(version)
    build_dists()
    dists = get_dists()
    if not long_description_is_ok(dists):
        sys.exit("Long description not ok.")
    if dryRun:
        sys.exit(f"This was a dry run for version {version}.")
    upload_dists(dists)
    push_released_tag(version)


def repo_is_dirty():
    try:
        git("diff", "--quiet")
        return False
    except ProcessExecutionError as e:
        if e.retcode != 1:
            raise
        return True


def tidy_up():
    for path in [DIST_PATH, PROJECT_ROOT_PATH / "build"]:
        if path.exists():
            shutil.rmtree(path)


def tag_repo(version):
    git("tag", version)


def build_dists():
    python = local["python"]
    python("setup.py", "sdist", "bdist_wheel")


def get_dists():
    distPath = PROJECT_ROOT_PATH / "dist"
    return list(str(dist) for dist in distPath.glob("*"))


def long_description_is_ok(dists):
    log.info(f"check {dists}")
    return not twine_check(dists)


def upload_dists(dists):
    settings = twine_Settings()
    twine_upload(settings, dists)


def push_released_tag(version):
    git("push", "origin", version)


if __name__ == "__main__":
    sys.exit(main())



"""High level build functionality bringing it all together."""
import logging
import os
import tempfile
import time
from pathlib import Path
from pprint import pformat

from i3configger import base, config, context, exc, ipc, message, partials

log = logging.getLogger(__name__)


def build_all():
    cnf = config.I3configgerConfig()
    log.info(f"start building from {cnf.partialsPath}")
    prts = partials.create(cnf.partialsPath)
    msg = message.Messenger(cnf.messagesPath, prts).fetch_messages()
    cnf.payload = context.merge(cnf.payload, msg[message.CMD.SHADOW])
    pathContentsMap = generate_contents(cnf, prts, msg)
    check_config(pathContentsMap[cnf.targetPath])
    persist_results(pathContentsMap)
    ipc.communicate(refresh=True)
    log.info("build done")


def generate_contents(cnf: config.I3configgerConfig, prts, msg):
    selectorMap = msg[message.CMD.SELECT]
    setMap = msg[message.CMD.SET]
    pathContentsMap = {}
    barTargets = cnf.get_bar_targets()
    selected = partials.select(prts, selectorMap, excludes=[base.I3BAR])
    ctx = context.process(selected + [setMap])
    log.debug(f"main context:\n{pformat(ctx)}")
    mainContent = generate_main_content(cnf.partialsPath, selected, ctx)
    for barName, barCnf in barTargets.items():
        barCnf["id"] = barName
        log.debug(f"bar {barName} config:\n{pformat(barCnf)}")
        extendedContext = context.process([ctx, barCnf])
        mainContent += f"\n{get_bar_setting(barCnf, prts, extendedContext)}"
        i3barFileContent = generate_i3bar_content(
            prts, barCnf["select"], extendedContext
        )
        if i3barFileContent:
            filename = f"{base.I3BAR}.{barCnf['select']}{base.SUFFIX}"
            dst = Path(barCnf["target"]) / filename
            pathContentsMap[dst] = i3barFileContent
    pathContentsMap[cnf.targetPath] = mainContent.rstrip("\n") + "\n"
    return pathContentsMap


def make_header(partialsPath):
    strPath = str(partialsPath)
    parts = strPath.split(os.getenv("HOME"))
    if len(parts) > 1:
        strPath = "~" + parts[-1]
    msg = f"# Built from {strPath} by i3configger ({time.asctime()}) #"
    sep = "#" * len(msg)
    return f"{sep}\n{msg}\n{sep}\n"


def generate_main_content(partialsPath, selected, ctx):
    out = [make_header(partialsPath)]
    for prt in selected:
        content = prt.get_pruned_content()
        if content:
            out.append(content)
    return context.substitute("\n".join(out), ctx).rstrip("\n") + "\n\n"


def get_bar_setting(barCnf, prts, ctx):
    tpl = partials.find(prts, base.I3BAR, barCnf["template"])
    assert isinstance(tpl, partials.Partial), tpl
    tpl.name = f"{tpl.name} [id: {barCnf['id']}]"
    return context.substitute(tpl.get_pruned_content(), ctx).rstrip("\n")


def generate_i3bar_content(prts, selectValue, ctx):
    prt = partials.find(prts, base.I3BAR, selectValue)
    if not prt:
        raise exc.ConfigError(
            f"[IGNORE] no status config named "
            f"{base.I3BAR}.{selectValue}{base.SUFFIX}"
        )
    assert isinstance(prt, partials.Partial), prt
    content = context.substitute(prt.get_pruned_content(), ctx)
    return content.rstrip("\n") + "\n"


def check_config(content):
    tmpPath = Path(tempfile.gettempdir()) / "i3config_check"
    tmpPath.write_text(content)
    errorReport = ipc.I3.get_config_error_report(tmpPath)
    if errorReport:
        msg = (
            f"FATAL: config not changed due to errors. "
            f"Broken config is at {tmpPath}"
        )
        report = f"config:\n{content}\n\nerrors:\n{errorReport}"
        ipc.communicate(msg, urgency="normal")
        raise exc.ConfigError(f"{msg}\n{report}")


def persist_results(pathContentsMap):
    for path, content in pathContentsMap.items():
        backupPath = Path(str(path) + ".bak")
        if path.exists() and not backupPath.exists():
            path.rename(backupPath)
        path.write_text(content)
